#!/usr/bin/env python3
# -*- coding: utf-8; py-indent-offset: 4 -*-


#  +-- bmSearch ------------------------------------------------------------------+
#  |                                                                              |
#  |  _                 _____                          _                          |
#  | | |               / ____|                        | |                         | 
#  | | |___  _ __ ___ | |____    ___  ____   _ _  ___ | |__                       |
#  | |  _  \| '_ ` _ \ \___  \  / _ \|___ \ | `_|/ __||  _  \                     | 
#  | | |_| || | | | | |____|  ||  __// __  || | | |__ | | | |                     |
#  | |_____/|_| |_| |_|______/  \___|\____/ |_|  \___||_| |_|                     |
#  |                                                                              |
#  |                                                                              |
#  +------------------------------------------------------------------------------+
#
# * ----------------------------------------------------------------------------------------*
# * @author  : Pascal SALAUN <pascal.salaun-AT-bm-monitor.org>                              *
# * @Website : http://www.bm-monitor.org <http://www.bm-monitor.org>                        *
# * @license : GNU Affero General Public License Version 3 <http://www.gnu.org/licenses>    *
# *                                                                                         *
# * bmSearch is free software;  you can redistribute it and/or modify it under the          *
# *  terms of the  license mentionned above.                                                *
# * bmSearch is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY    * 
# * ----------------------------------------------------------------------------------------*

# Native Python Lib/Class
from datetime import datetime, date, timedelta

class bmDate:
    
    def __init__(self):
        """
            __init__ : 
                generate a dict containing dates formated as described under : 

                    dates[yesterdayPostfixLiteral] = %Y-%m-%d, where yesterdayPostfixLiteral = "%b %_d", ie dates['Jul 07'] = "2017-07-06'
                    dates[todayPostfixLiteral] = %Y-%m-%d, where todayPostfixLiteral = "%b %_d", ie dates['Jul 07'] = "2017-07-07'
                    dates['yesterdayNumELK'] = %Y.%m.%d, ie dates['yesterdayNumELK'] = '2017.07.06'
                    dates['todayNumELK'] = %Y.%m.%d, ie dates['yesterdayNumELK'] = '2017.07.07'
        """
                
        self.dates={}
                
        today                   = date.today()
        todayNum                = today.strftime("%Y-%m-%d")
        todayNumELK             = today.strftime("%Y.%m.%d")
        todayPostfixLiteral     = today.strftime("%b %_d")

        yesterday               = today - timedelta(days=1)
        yesterdayNum            = yesterday.strftime("%Y-%m-%d")
        yesterdayNumELK         = yesterday.strftime("%Y.%m.%d")
        yesterdayPostfixLiteral = yesterday.strftime("%b %_d")

                
        self.dates[yesterdayPostfixLiteral] = yesterdayNum 
        self.dates[todayPostfixLiteral]     = todayNum
        self.dates['yesterdayNumELK']       = yesterdayNumELK
        self.dates['todayNumELK']           = todayNumELK


    def getDates(self):
        """
            getDates :
                Return dates dict
        """
        return self.dates


    def setDates(self, setDate):
        today                   = date.fromisoformat(setDate)
        todayNum                = today.strftime("%Y-%m-%d")
        todayNumELK             = today.strftime("%Y.%m.%d")
        todayPostfixLiteral     = today.strftime("%b %_d")

        yesterday               = today - timedelta(days=1)
        yesterdayNum            = yesterday.strftime("%Y-%m-%d")
        yesterdayNumELK         = yesterday.strftime("%Y.%m.%d")
        yesterdayPostfixLiteral = yesterday.strftime("%b %_d")

                
        self.dates[yesterdayPostfixLiteral] = yesterdayNum 
        self.dates[todayPostfixLiteral]     = todayNum
        self.dates['yesterdayNumELK']       = yesterdayNumELK
        self.dates['todayNumELK']           = todayNumELK
        
        


    # OBSOLETE
    def getCyrusUpdatePoolDates(self,**kwargs):
        """
            getCyrusUpdatePoolDates:
                return a list of pre-formated date as "%b %_d %H:%M"
                The number of dates are hard-coded as n=5 for 5 minutes

        """
        if "setDate"  in kwargs.keys():
            now =  date.fromisoformat(kwargs['setDate'])
        else:
            now = datetime.today()
        """
            n : number of  minutes. Must be the same as set in "/etc/cron.d/bmSearch" (cf updateCyrusInfo.py)
        """
        n=5
        dateList = []
        while n > 0 :
            before = now - timedelta(minutes=n)
            dateList.append(before.strftime("%b %_d %H:%M"))
            n -=1


        return dateList
                
                
                
                
                
                
