#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
 * ------------------------------------------------------------------------
 * bmSearch
 * ------------------------------------------------------------------------
 * @license : GNU Affero General Public License Version 3 <http://www.gnu.org/licenses>
 * @author : Pascal SALAUN 
 * @Website : http://www.bm-stats.org <http://www.bm-stats.org>
 * ------------------------------------------------------------------------
"""

# Native Python Lib/Class 
from collections import defaultdict
from elasticsearch import Elasticsearch, helpers
import json



# bmSearch Class
import config as CFG

from logger import *
backoffice = logger().setLogger().getChild('_elastic')


class elastic:
    def __init__(self):
        """
            __init__ :
                Set default headers
                Set a (dict) dates from bmDate
        """
        
        self.params = defaultdict(dict)
        
        self.APIURL           = CFG.APIURL
        
        if hasattr(CFG,'user') and hasattr(CFG,'password'):
            self.params['http_auth'] = (CFG.user, CFG.password)
            
        if hasattr(CFG,'verifyPeer'):
            self.params['verify'] = CFG.verifyPeer
        
        if hasattr(CFG,'serverCert'):
            self.params['ca_certs'] = CFG.serverCert

        if 'https' in self.APIURL: 
            self.params['use_ssl'] = True
        
        
        
    def setParam(self, **kwargs):
        if "user" in kwargs.keys() and  "password" in kwargs.keys():
            self.params['http_auth'] = (kwargs['user'],kwargs['password'])
        
        if "verifyPeer" in kwargs.keys():
            self.params['verify'] = kwargs['verifyPeer']
        
        if "serverCert" in kwargs.keys():
            self.params['ca_certs'] = kwargs['serverCert']
        
        
        
    def initSession(self): 
        self.es = Elasticsearch([self.APIURL], **self.params)
            
    
    def _bulk(self, *args):
        (_index, _type, _data)= args[0]
        actions = [
            {
                '_index': _index,
                '_type': _type,
                '_source': _dict,
                }
                for _dict in _data
            ]  
            
        helpers.bulk(self.es, actions)
