#!/usr/bin/env python3
# -*- coding: utf-8; py-indent-offset: 4 -*-


#  +-- bmSearch ------------------------------------------------------------------+
#  |                                                                              |
#  |  _                 _____                          _                          |
#  | | |               / ____|                        | |                         | 
#  | | |___  _ __ ___ | |____    ___  ____   _ _  ___ | |__                       |
#  | |  _  \| '_ ` _ \ \___  \  / _ \|___ \ | `_|/ __||  _  \                     | 
#  | | |_| || | | | | |____|  ||  __// __  || | | |__ | | | |                     |
#  | |_____/|_| |_| |_|______/  \___|\____/ |_|  \___||_| |_|                     |
#  |                                                                              |
#  |                                                                              |
#  +------------------------------------------------------------------------------+
#
# * ----------------------------------------------------------------------------------------*
# * @author  : Pascal SALAUN <pascal.salaun-AT-bm-monitor.org>                              *
# * @Website : http://www.bm-monitor.org <http://www.bm-monitor.org>                        *
# * @license : GNU Affero General Public License Version 3 <http://www.gnu.org/licenses>    *
# *                                                                                         *
# * bmSearch is free software;  you can redistribute it and/or modify it under the          *
# *  terms of the  license mentionned above.                                                *
# * bmSearch is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY    * 
# * ----------------------------------------------------------------------------------------*

#NativePythonLib/Class
from collections import defaultdict
try:
    import cStringIO
    io_method=cStringIO.StringIO
except ImportError:
    from io import StringIO
    io_method=StringIO
    
from datetime import *
from email.header import Header, decode_header, make_header
import hashlib
import json
import magic
import os
import re
import subprocess
import sys
import tarfile
import time


#bmSearchClass
from bmObject import *
from bmDate import *
import config as CFG
from fileMIMEType import *
from regex import *
from requestToELK import *
from sqlQuery import *

from logger import *
backoffice = logger().setLogger().getChild('mailLogParser')


class mailLogParser:
    """
    mailLogParser:
    Hastoopen/streammail.logfile,
    parseeachlinetofeedadictwithpertinentvalues
    andtry(ifkwargsset)topublishtheminELKplatform
    """

    def __init__(self):
        """
        __init__:
                constructoracceptkwargs

                **kwargs
                file:pathtofiletoparse.Icanbeplaintext,gziporbzip2compressed.Bydefaultwhatconfiguredin/etc/bmSearch/config.py

                method:howtoreadthefile,2possibilities
                -"file":readthefileandexitafterreachendfile.Bydefault.
                -"stream":connecttothefileandstreamdata.Noexitwhileprogrammrunning

                action:tellwhattodo,2cases
                -"storeToELK":tostoreinElasticsearchinstanceconfiguredin/etc/bmSearch/config.py
                -"updateToELK":toupdateinElasticsearchinstancedata(deliveredRecip)

        """
        # Global vars
        self.parseAntivirusLine=False
        self.pushCyrusDeliveryStatus=False
        self.pushCyrusLoginStatus=False
        self.pushDSN2Status=False
        self.pushDSN4Status=False
        self.pushDSN5Status=False
        self.pushRoutedLines=False
        
        
        # Overload from config file if set
        if hasattr(CFG,'parseAntivirusLine'):
            self.parseAntivirusLine = CFG.parseAntivirusLine

        if hasattr(CFG,'pushCyrusDeliveryStatus'):
            self.pushCyrusDeliveryStatus = CFG.pushCyrusDeliveryStatus
        if hasattr(CFG,'pushCyrusLoginStatus'):
            self.pushCyrusLoginStatus = CFG.pushCyrusLoginStatus
        
        if hasattr(CFG,'pushDSN2Status'):
            self.pushDSN2Status = CFG.pushDSN2Status
        if hasattr(CFG,'pushDSN4Status'):
            self.pushDSN4Status = CFG.pushDSN4Status
        if hasattr(CFG,'pushDSN5Status'):
            self.pushDSN5Status = CFG.pushDSN5Status
            
        if hasattr(CFG,'pushRoutedLines'):
            self.pushRoutedLines = CFG.pushRoutedLines

        # Main data dict
        self.data                 = defaultdict(dict)
        self.surveyMessageid      = []
        self.surveyQueuid         = []
        self.queueidFromMessageid = defaultdict(dict)
        
        self.bmObjectDSN4          = [] # Use to push obj to ELK when obj is older than n minutes, because no close line found, e.g. in case of DSN 4.

        self.statusHashLineList = []

        # Only for Antivirus case
        self.addonInfo = defaultdict(dict)

        #Setatupleoffiletypestoinit(cfconnexion)
        self.theseFileTypes = CFG.fileTypeTupl
        
        

        
        """
        Postfix:theselinesareaboutpostfixactivityforonemail
        """

        self.smtpdLine       = smtpdLine
        self.cleanupLine     = cleanupLine
        self.filenameLine    = filenameLine
        self.bodyDateLine    = bodyDateLine
        self.bodyFromLine    = bodyFromLine
        self.bodySubjectLine = bodySubjectLine
        self.userAgentLine   = userAgentLine
        self.qmgrLine        = qmgrLine
        self.smtpLineDSNAll  = smtpLineDSNAll
        self.smtpLineDSN2    = smtpLineDSN2
        self.smtpLineDSN4    = smtpLineDSN4
        self.smtpLineDSN5    = smtpLineDSN5
        self.smtpRemoved     = smtpRemoved

        """
        Cyrus/IMAP:thislineisrelativetoadéliverysuccess
        """
        self.cyrusDelivered = cyrusDelivered
        self.cyrusLogin     = cyrusLogin

        """
        DATE
        """
        self.dates = bmDate().getDates()
        self.setDate = None
        self.deltaMicroStp = 0
        self.deltaPrevSec = "00"

        """
        cfhttps://stackoverflow.com/questions/7331351/python-email-header-decoding-utf-8AbhishekGurjar
        """
        self.pat2 = re.compile(r'(([^=]*)=\?([^\?]*)\?([BbQq])\?([^\?]*)\?=([^=]*))',re.IGNORECASE)

        """
        Session Pool connection
        """
        self.s = requestToELK()
        self.s.initSession()



    def setParam(self,**kwargs):
        """
            setParam:
                overload params with these in **kwargs.
                    Paramsavailableare:
                    file
                    action
                    method
                    dbStore
                    dbtemp
                    dbStoreMaxDays
                    timeInterval
                    Fordetailshavealookin/etc/bmSearch/config.py
        """
        self.maillog = CFG.maillog
        if "file" in kwargs.keys():
            self.maillog = kwargs['file']

        self.action = ""
        if "action" in kwargs.keys():
            self.action=kwargs['action']

        self.method = "file"
        if "method" in kwargs.keys():
            self.method = kwargs['method']

        self.dbStore = CFG.dbStore
        if "dbStore" in kwargs.keys():
            self.dbStore = kwargs['dbStore']

        self.dbtemp = CFG.dbtemp
        if "dbtemp" in kwargs.keys():
            self.dbStore = kwargs['dbtemp']

        self.dbStoreMaxDays = CFG.dbStoreMaxDays
        if "dbStoreMaxDays" in kwargs.keys():
            self.dbStoreMaxDays=kwargs['dbStoreMaxDays']

        # Will be removed soon
        if hasattr(CFG,'timeInterval'):
            self.timeInterval = CFG.timeInterval
        if "timeInterval" in kwargs.keys():
            self.timeInterval = kwargs['timeInterval']


        if "setDate"  in kwargs.keys():
            self.setDate = kwargs['setDate']
            
            d = bmDate()
            d.setDates(kwargs['setDate'])
            self.dates = d.getDates()

        if 'parseAntivirusLine' in kwargs.keys():
            self.parseAntivirusLine = kwargs['parseAntivirusLine']

        if 'pushCyrusDeliveryStatus' in kwargs.keys():
            self.pushCyrusDeliveryStatus = kwargs['pushCyrusDeliveryStatus']
            
        if 'pushCyrusLoginStatus' in kwargs.keys():
            self.pushCyrusLoginStatus = kwargs['pushCyrusLoginStatus']
        
        if 'pushDSN2Status' in kwargs.keys():
            self.pushDSN2Status = kwargs['pushDSN2Status']
            
        if 'pushDSN4Status' in kwargs.keys():
            self.pushDSN4Status = kwargs['pushDSN4Status']
            
        if 'pushDSN5Status' in kwargs.keys():
            self.pushDSN5Status = kwargs['pushDSN5Status']
            
        if 'pushRoutedLines' in kwargs.keys():
            self.pushRoutedLines = kwargs['pushRoutedLines']
            
        self.s.setParam(**kwargs)

    def getStreamContent(self):
        """
            getStreamContent:
                Connect to a file and stream its content(as a tail). Its name is from config or bmSearch(main) kwargs "file" (seebmSearch).
                Setvar"content"withline(s)
                Breakonlywhenmainbreaks
        """
        try:
            stream = open(self.maillog)

            interval=1.0

            while True:
                where = stream.tell()
                line = stream.readline()

                if not line:
                    time.sleep(interval)
                    stream.seek(where)
                else:
                    yield line

        except Exception as error:
            backoffice.critical(error)



    def getFileContent(self):
        """
            getFileContent:
            Read a whole file. Its name is from config or bmSearch(main) kwargs "file" (seebmSearch).
                Set var "content" with line(s)

        """
        maillog = self.maillog.encode('utf8')

        compressionCase = {"text/plain":"plaintext","application/x-gzip":"r:gz","application/gzip":"r:gz","application/x-bzip2":"r:bz2"}
        t = fileMIMEType()
        ftype = t.getType(maillog)

        if compressionCase[ftype] == "plaintext":
            try:
                stream = open(maillog)
                while True:
                    yield stream.readline()

            except Exception as error:
                backoffice.critical(error)
        else:
            try:
                ft = tarfile.open(maillog,compressionCase[ftype])
                member0 = ft.getnames()[0]
                f = ft.extractfile(member0)
                c = f.readlines()
                ft.close()
                
                for line in c:
                    yield line.decode('utf8')

            except tarfile.ReadError:
                p = subprocess.Popen(["zcat",maillog],stdout=subprocess.PIPE)
                c = io_method(p.communicate()[0])
                assertp.returncode == 0
                
                for line in c:
                    yield line.decode('utf8')

            except Exception as error:
                backoffice.critical(error)



    def parseContent(self):
        """
            parseContent:
                read line from getFileContent(=>method='file') or getStreamContent (=>method='stream') and extract data
                If bmSearch (main) kwargs "action" isset and has value:
                    -"storeToELK",try to post data to an Elasticsearch instance
                    -"updateToELK",try to update an Elasticsearch instance data(deliveredRecip)
        """

        """
        OpenSQLITEDB
        """
        if self.method == "file":
            db = self.dbtemp
        if self.method == "stream":
            db = self.dbStore

        self.cDb = sqlQuery(db)
        self.cDb.initDb()
        self.cDb.updateTrigger(self.dbStoreMaxDays)
        self.cDb.updateStatusTrigger()
        self.surveyMessageid = self.cDb.loadMessageid()

        # Preprared case 
        self.statusHashLineList = self.cDb.loadStatusHashLine()

        if self.method == "file":
            for line in self.getFileContent():
                self.parseLine(line)

        if self.method == "stream":
            for line in self.getStreamContent():
                self.parseLine(line)



    def decodev2(self,bodySubject):
        """
            decodev2:
                return email subject in pretty view
                Args:
                    bodySubject; the email subject entry
        """
        h = make_header(decode_header(bodySubject))
        return str(h).replace(' ? ','')


    def getDeltaMicroStp(self,hour):
        """
        getDeltaMicroStp:
            Used to order entries when many mail treated in same second
            Return an int to sum to timestamp 
        Args:
            (str) hour : the hour as HH:MM:SS
        
        """
        sec = hour.split(":")[2]
        
        if sec == self.deltaPrevSec:
            self.deltaMicroStp  += 1
        else:
            self.deltaMicroStp = 0
            
        self.deltaPrevSec = sec
        return self.deltaMicroStp
        
        

    def parseLine(self,line):
        """
        parseLine:
            read line and extract data
            If bmSearch(main) kwargs "action" is set and has value:
                -"storeToELK",try to post data to an Elasticsearch instance
                -"updateToELK",try to update an Elasticsearch instance data(deliveredRecip)
        """
        thisLine = line.rstrip()
        
        # Connection
        smtpdLineEntries = self.smtpdLine.match(thisLine)
        if smtpdLineEntries is not None:
            # try to update some datas (antivirus)
            if self.setDate is not None:
                date = self.setDate
            else:
                date = time.strftime('%Y-%m-%d',time.localtime())
                
            try:
                date=self.dates[smtpdLineEntries.group('date')]
            except:
                pass

            hour    = smtpdLineEntries.group('hour')
            timestp = int(datetime.strptime(date+""+hour,'%Y-%m-%d%H:%M:%S').strftime("%s"))*1000 + self.getDeltaMicroStp(hour)
            queueid = smtpdLineEntries.group('queueid')
            server  = smtpdLineEntries.group('server').split(",")[0].replace("[",":").replace("]","")

            self.data[queueid] = bmObject().initMsgObj()
            self.data[queueid]['@timestamp'] = timestp
            self.data[queueid]['date']       = date + " " +  hour
            self.data[queueid]['hour']       = hour
            self.data[queueid]['server']     = server

            #ElasticEntries
            self.data[queueid]['source'] = self.maillog
            self.data[queueid]['beat']['hostname'] = CFG.HOSTNAME
            self.data[queueid]['beat']['version']  = CFG.VERSION



            self.data[queueid]['fileType'] = defaultdict(dict)
            for thisType in self.theseFileTypes:
                self.data[queueid]['fileType'][thisType] = 0


        # Message-id
        cleanupLineEntries=self.cleanupLine.match(thisLine)
        if cleanupLineEntries is not None:            
            if self.setDate is not None:
                date = self.setDate
            else:
                date = time.strftime('%Y-%m-%d',time.localtime())
                
            try:
                date=self.dates[cleanupLineEntries.group('date')]

            except:
                pass
            hour      = cleanupLineEntries.group('hour')
            server    = cleanupLineEntries.group('server')
            queueid   = cleanupLineEntries.group('queueid')
            #messageid = cleanupLineEntries.group('messageid').replace("<","").replace(">","")
            messageid = re.sub(r'[^a-zA-Z0-9_\-\.@]+', '.', cleanupLineEntries.group('messageid').replace("<","").replace(">","")) 
            timestp = int(datetime.strptime(date+""+hour,'%Y-%m-%d%H:%M:%S').strftime("%s"))*1000 + self.getDeltaMicroStp(hour)
            
            # Because empty message-id, so have to define one
            if len(messageid) == 0:
                messageid = queueid + "-waf-" + str(timestp) + "@empty.local"
            
            #Just in case of no connection line (case already seen while syslog over udp)
            if not queueid in self.data.keys():
                self.data[queueid] = bmObject().initMsgObj()
                self.data[queueid]['@timestamp'] = timestp
                self.data[queueid]['date']       = date + " " +  hour
                self.data[queueid]['hour']       = hour
                self.data[queueid]['server']     = server

                #ElasticEntries
                self.data[queueid]['source'] = self.maillog
                self.data[queueid]['beat']['hostname'] = CFG.HOSTNAME
                self.data[queueid]['beat']['version']  = CFG.VERSION

                self.data[queueid]['fileType'] = defaultdict(dict)
                for thisType in self.theseFileTypes:
                    self.data[queueid]['fileType'][thisType] = 0

            self.data[queueid]['msgid'] = messageid
                            

            #A hack because in case of delivery notification(from=<>),it could not have an"postfix/smtpd" line
            if 'recip' in self.data[queueid].keys():
                pass
            else:
                self.data[queueid]['recip']=[]
               
            # Store to ELK line about where (server) mail passed through   
            if self.pushRoutedLines: 
                # Track each lines for having a knowledge of where mail passed through 
                ## Test if this line is already parsed 
                hashLine = hashlib.md5(thisLine.encode('ascii')).hexdigest()
                #if hashLine in self.statusHashLineList:         
                if self.cDb.verifyIfStatusHashLineExists(hashLine):
                    pass
                else:             
                    passedThrough =  bmObject().passedThroughObj()
                    passedThrough['@timestamp'] = self.data[queueid]['@timestamp']
                    passedThrough['date']       = self.data[queueid]['date']
                    passedThrough['hour']       = self.data[queueid]['hour']

                    #ElasticEntries
                    passedThrough['source'] = self.maillog
                    passedThrough['beat']['hostname'] = CFG.HOSTNAME
                    passedThrough['beat']['version']  = CFG.VERSION

                    passedThrough['server']  = server
                    passedThrough['msgid']   = messageid           
                    passedThrough['queueid'] = queueid
                    passedThrough['log']     = thisLine

                    if self.action=="storeToELK":
                        self.s.sendDataRoutedBy(passedThrough)
                        
                        
                    tupl=(hashLine, passedThrough['@timestamp'])
                    self.cDb.addStatusEntry(tupl)
                    passedThrough = defaultdict(dict)                
                    #self.statusHashLineList.append(hashLine)
                
            # Delete from general data dict if already tracked/parsed, otherwise continue parsing mail logs
            if messageid in self.surveyMessageid:
                del self.data[queueid]
            else:
                self.surveyMessageid.append(messageid)
                self.surveyQueuid.append(queueid)
                self.queueidFromMessageid[messageid] = queueid

            if '@timestamp' in self.data[queueid].keys():
                timestamp = self.data[queueid]['@timestamp']/1000
            else:
                timestamp = int(time.time())

        ## ANTIVIRUS
        # Analyze and stuff if configured/set. By default False
        if self.parseAntivirusLine:
            # KasperkyLabs
            klmsLineEntries = klmsLine.match(thisLine.replace("\"",""))
            if klmsLineEntries is not None:
                messageid = re.sub(r'[^a-zA-Z0-9_\-\.@]+', '.', klmsLineEntries.group('messageid').replace("<","").replace(">",""))
                status    = klmsLineEntries.group('status')
                status   += ",antivirus-status:"+klmsLineEntries.group('avstatus')
                status   += ",antiphishing-status:"+klmsLineEntries.group('apstatus')
                status   += ",antispam-status:"+klmsLineEntries.group('asstatus')

                if messageid in self.queueidFromMessageid.keys():
                    queueid = self.queueidFromMessageid[messageid]
                    self.data[queueid]['antivirus']=status

                    try:
                        del self.queueidFromMessageid[messageid]
                    except KeyError:
                        pass


            # Amavis-New
            amavisLineEntries = amavisLine.match(thisLine)
            if amavisLineEntries is not None:
                messageid = re.sub(r'[^a-zA-Z0-9_\-\.@]+', '.', amavisLineEntries.group('messageid').replace("<","").replace(">",""))
                status    = amavisLineEntries.group('status')

                if messageid in self.queueidFromMessageid.keys():
                    queueid = self.queueidFromMessageid[messageid]
                    
                    self.data[queueid]['antivirus'] = status

                    try:
                        del self.queueidFromMessageid[messageid]
                    except KeyError:
                        pass


        # Filenameifcontroled
        filenameLineEntries=self.filenameLine.match(thisLine)
        if filenameLineEntries is not None:
            queueid  = filenameLineEntries.group('queueid')
            filename = filenameLineEntries.group('filename').replace("\"","").split(";")[0]

            if queueid in self.data and len(queueid):

                if 'filename' not in self.data[queueid]:
                    self.data[queueid]['filename'] = []

                if 'fileType' not in self.data[queueid]:
                    try:
                        self.data[queueid]['fileType'] = defaultdict(dict)
                    except:
                        pass

                fileType = filename.lower().split('.')[-1].replace("jpg","jpeg")

                if filename not in self.data[queueid]['filename']:
                    self.data[queueid]['filename'].append(filename)

                    try:
                        self.data[queueid]['fileType'][fileType] +=1
                    except:
                        self.data[queueid]['fileType'][fileType] =0
                        self.data[queueid]['fileType'][fileType] +=1


        ## Specific body message checks(see/etc/postfix/BMSEARCH_header_checks)

        # For these header controls we have to set the self.data[queueid] in every case, because lines
        # are written by cleanup,before the message-id
        # bodyDate if controled
        bodyDateLineEntries=self.bodyDateLine.match(thisLine)
        if bodyDateLineEntries is not None:
            queueid  = bodyDateLineEntries.group('queueid')
            bodyDate = bodyDateLineEntries.group('bodyDate')

            if queueid in self.data:
                self.data[queueid]['bodyDate'] = bodyDate


        # userAgent if controled
        userAgentLineEntries=self.userAgentLine.match(thisLine)
        if userAgentLineEntries is not None:
            queueid   = userAgentLineEntries.group('queueid')
            userAgent = userAgentLineEntries.group('userAgent')

            if queueid in self.data:
                self.data[queueid]['userAgent'] = userAgent
                
                
        # bodyFrom if controled
        bodyFromLineEntries=self.bodyFromLine.match(thisLine)
        if bodyFromLineEntries is not None:
            queueid  = bodyFromLineEntries.group('queueid')
            bodyFrom = bodyFromLineEntries.group('bodyFrom')
    
            if queueid in self.data:
                self.data[queueid]['bodyFrom']       = bodyFrom.lower().split('<')[-1].replace(">","")
                self.data[queueid]['bodyFromDomain'] = bodyFrom.lower().split('@')[-1].replace(">","")


        # bodySubject if controled
        bodySubjectLineEntries=self.bodySubjectLine.match(thisLine)
        if bodySubjectLineEntries is not None:
            queueid     = bodySubjectLineEntries.group('queueid')
            bodySubject = bodySubjectLineEntries.group('bodySubject')

            if queueid in self.data:
                try:
                    self.data[queueid]['bodySubject'] = self.decodev2(bodySubject)
                except:
                    self.data[queueid]['bodySubject'] = bodySubject

        ## Back to general info.

        #Sender,size,nrcpt
        qmgrLineEntries=self.qmgrLine.match(thisLine)
        if qmgrLineEntries is not None:
            queueid  = qmgrLineEntries.group('queueid')
            thisFrom = qmgrLineEntries.group('from')
            size     = qmgrLineEntries.group('size')
            nrcpt    = qmgrLineEntries.group('nrcpt')

            if queueid in self.surveyQueuid:    
                if len(thisFrom):
                    pass
                else:
                    thisFrom='notification@none'

                self.data[queueid]['size']       = size
                self.data[queueid]['nrcpt']      = nrcpt
                self.data[queueid]['from']       = thisFrom.lower()
                self.data[queueid]['fromDomain'] = thisFrom.lower().split('@')[-1].replace(">","")

        
        # Get all Recip name smtpLineDSNAll
        smtpLineDSNAllEntries=self.smtpLineDSNAll.match(thisLine)
        if smtpLineDSNAllEntries is not None:
            queueid = queueidas = smtpLineDSNAllEntries.group('queueid')
            recip = orig = smtpLineDSNAllEntries.group('recip').lower()
            relay = smtpLineDSNAllEntries.group('relay')
            
            if queueid in self.surveyQueuid:
                if smtpLineDSNAllEntries.group('orig') is not None:
                    orig      = smtpLineDSNAllEntries.group('orig').lower()
                
                if smtpLineDSNAllEntries.group('queueidas') is not None:
                    queueidas = smtpLineDSNAllEntries.group('queueidas')

                if queueid == queueidas:
                    routing = "internal"
                else:
                    routing = "external"

                Recip = defaultdict(dict)         
                Recip['address']   = recip
                Recip['domain']    = recip.lower().split('@')[-1]
                Recip['routing']   = routing
                Recip['orig']      = orig
                Recip['relay']     = relay
                Recip['queueid']   = queueid
                Recip['queueidas'] = queueidas



                if 'recip' in self.data[queueid]:
                    self.data[queueid]['recip'].append(Recip)
                else:
                    self.data[queueid]['recip'] = []
                    self.data[queueid]['recip'].append(Recip)

                del Recip
            
                ## Remove duplicate
                #rcp = self.data[queueid]['recip']
                #self.data[queueid]['recip'] = [ dict(t) for t in {tuple(d.items()) for d in rcp} ]
                
                # Because of DSN 4 where no remove line.
                # So have to store object 
                if queueid in self.bmObjectDSN4:
                    timestp = self.data[queueid]['@timestamp']/1000
                    queuidObj = json.dumps(self.data[queueid])
                    qtupl = (queueid, queuidObj, timestp)
                    self.cDb.upsertQueueObjEntry(qtupl)  
             
        
        # When Postfix delivery is OK
        if self.pushDSN2Status:                
            smtpLineDSN2Entries = self.smtpLineDSN2.match(thisLine)
            if smtpLineDSN2Entries is not None:
                hashLine = hashlib.md5(thisLine.encode('ascii')).hexdigest()

                # Test if this line is already parsed
            
                # Preprared line in case of load to memory
                #if hashLine in self.statusHashLineList:
                if self.cDb.verifyIfStatusHashLineExists(hashLine):
                    pass
                else:
                    
                    queueid = smtpLineDSN2Entries.group('queueid')
                    recip   = smtpLineDSN2Entries.group('recip')
                    hour    = smtpLineDSN2Entries.group('hour')
                   
                    if queueid in self.surveyQueuid:
                        if self.setDate is not None:
                            date = self.setDate
                        else:
                            date = time.strftime('%Y-%m-%d',time.localtime())

                        try: 
                            date=self.dates[smtpLineDSN2Entries.group('date')]
                        except:
                            pass
                
                        if smtpLineDSN2Entries.group('queueidas'):
                            queueidas = smtpLineDSN2Entries.group('queueidas')
            
                        timestp = int(datetime.strptime(date+""+hour,'%Y-%m-%d%H:%M:%S').strftime("%s"))*1000 + self.getDeltaMicroStp(hour)

                        # Feed status table with this new hash for next time bmSearch will re-parsed this this
                        tupl=(hashLine, timestp)
                        self.cDb.addStatusEntry(tupl)
                        #self.statusHashLineList.append(hashLine)

                        dsn = bmObject().dsnObj()
                        dsn['@timestamp'] = timestp
                        dsn['date']       = date + " " +  hour
                        dsn['hour']       = hour

                        #ElasticEntries
                        dsn['source'] = self.maillog
                        dsn['beat']['hostname'] = CFG.HOSTNAME
                        dsn['beat']['version']  = CFG.VERSION

                        dsn['recip']   = recip
                        dsn['domain'] = recip.lower().split('@')[-1].replace(">","")
                        
                        dsn['dsn']['queueid'] = queueid
                        dsn['dsn']['status'] = "DSN 2"
                        dsn['dsn']['log']    = thisLine

                        if self.action=="storeToELK":
                            self.s.sendDataStatus(dsn)
                            
                        # After post data to ELK 
                        try:
                            del dsn
                        except  KeyError:
                            pass

        
        #When Postfix delivery is deferred (ie mailbox is over quota)
        smtpLineDSN4Entries = self.smtpLineDSN4.match(thisLine)
        if smtpLineDSN4Entries is not None:
            hashLine = hashlib.md5(thisLine.encode('ascii')).hexdigest()

            # Test if this line is already parsed
            
            #if hashLine in self.statusHashLineList:
            if self.cDb.verifyIfStatusHashLineExists(hashLine):
                pass
            else:
                queueid = smtpLineDSN4Entries.group('queueid')
                recip   = smtpLineDSN4Entries.group('recip')
                hour    = smtpLineDSN4Entries.group('hour')
                
                if queueid in self.surveyQueuid:
                    if self.setDate is not None:
                        date = self.setDate
                    else:
                        date = time.strftime('%Y-%m-%d',time.localtime())

                    try:
                        date=self.dates[smtpLineDSN4Entries.group('date')]
                    except:
                        pass
                    
                    timestp = int(datetime.strptime(date+""+hour,'%Y-%m-%d%H:%M:%S').strftime("%s"))*1000 + self.getDeltaMicroStp(hour)
                    
                    
                    if self.pushDSN4Status:
                        # Feed status table with this new hash for next time bmSearch will re-parsed this this
                        tupl=(hashLine, timestp)
                        self.cDb.addStatusEntry(tupl)
                        #self.statusHashLineList.append(hashLine)
                        
                        dsn = bmObject().dsnObj()
                        dsn['@timestamp'] = timestp
                        dsn['date']       = date + " " +  hour
                        dsn['hour']       = hour

                        #ElasticEntries
                        dsn['source'] = self.maillog
                        dsn['beat']['hostname'] = CFG.HOSTNAME
                        dsn['beat']['version']  = CFG.VERSION

                        dsn['recip']  = recip
                        dsn['domain'] = recip.lower().split('@')[-1].replace(">","")

                        dsn['dsn']['queueid'] = queueid
                        dsn['dsn']['status']  = "DSN 4"
                        dsn['dsn']['log']     = thisLine

                        if self.action=="storeToELK":
                            self.s.sendDataStatus(dsn)
                    
                        # After post data to ELK 
                        try:
                            del dsn
                        except  KeyError:
                            pass
                    
                        # Because of DSN 4 where no remove line.
                        # So have to store object 

                        queuidObj = json.dumps(self.data[queueid])
                        qtupl = (queueid, queuidObj, timestp/1000)
                        self.cDb.upsertQueueObjEntry(qtupl)  
                        self.bmObjectDSN4.append(queueid)
                


        
        # When Postfix delivery is bounced (ie the mailbox does not exist or the msg is older than 5 days)
        if self.pushDSN5Status:           
            smtpLineDSN5Entries=self.smtpLineDSN5.match(thisLine)
            if smtpLineDSN5Entries is not None:
                hashLine = hashlib.md5(thisLine.encode('ascii')).hexdigest()
                
                # Test if this line is already parsed
                #if hashLine in self.statusHashLineList:
                if self.cDb.verifyIfStatusHashLineExists(hashLine):
                    pass
                else:
                    
                    queueid = smtpLineDSN5Entries.group('queueid')
                    recip   = smtpLineDSN5Entries.group('recip')
                    hour    = smtpLineDSN5Entries.group('hour')
                    
                    if queueid in self.surveyQueuid:
                        if self.setDate is not None:
                            date = self.setDate
                        else:
                            date = time.strftime('%Y-%m-%d',time.localtime())

                        try:
                            date=self.dates[smtpLineDSN5Entries.group('date')]
                        except:
                            pass
                        
                        timestp = int(datetime.strptime(date+""+hour,'%Y-%m-%d%H:%M:%S').strftime("%s"))*1000 + self.getDeltaMicroStp(hour)
                    
                        # Feed status table with this new hash for next time bmSearch will re-parsed this this
                        tupl=(hashLine, timestp)
                        self.cDb.addStatusEntry(tupl)
                        #self.statusHashLineList.append(hashLine)
                        
                        
                        dsn = bmObject().dsnObj()
                        dsn['@timestamp'] = timestp
                        dsn['date']       = date + " " +  hour
                        dsn['hour']       = hour

                        #ElasticEntries
                        dsn['source'] = self.maillog
                        dsn['beat']['hostname'] = CFG.HOSTNAME
                        dsn['beat']['version']  = CFG.VERSION

                        dsn['recip']  = recip
                        dsn['domain'] = recip.lower().split('@')[-1].replace(">","")

                        dsn['dsn']['queueid'] = queueid
                        dsn['dsn']['status']  = "DSN 5"
                        dsn['dsn']['log']     = thisLine

                        
                        if self.action=="storeToELK":
                            self.s.sendDataStatus(dsn)

                        # After post data to ELK 
                        try:
                            del dsn
                        except  KeyError:
                            pass

        
        # Analyze and stuff if configured/set. By default False
        if self.pushCyrusDeliveryStatus: 
            # When Cyrus delivery msg to mailbox 
            cyrusDeliveredEntries = self.cyrusDelivered.match(thisLine)        
            if cyrusDeliveredEntries is not None:
                hashLine = hashlib.md5(thisLine.encode('ascii')).hexdigest()
                
                # Test if this line is already parsed
                #if hashLine in self.statusHashLineList:
                if self.cDb.verifyIfStatusHashLineExists(hashLine):
                    pass
                else:
                    
                    date      = self.dates[cyrusDeliveredEntries.group('date')]
                    hour      = cyrusDeliveredEntries.group('hour')
                    #messageid = cyrusDeliveredEntries.group('messageid').replace("<","").replace(">","")
                    messageid = re.sub(r'[^a-zA-Z0-9_\-\.@]+', '.', cyrusDeliveredEntries.group('messageid').replace("<","").replace(">",""))
                    domain    = cyrusDeliveredEntries.group('domain')
                    user      = cyrusDeliveredEntries.group('user')
                                
                    recip = user.replace("^", ".") + "@" + domain 

                    timestp = int(datetime.strptime(date+""+hour,'%Y-%m-%d%H:%M:%S').strftime("%s"))*1000 + self.getDeltaMicroStp(hour)
                    
                    # Feed status table with this new hash for next time bmSearch will re-parsed this this
                    tupl=(hashLine, timestp)
                    self.cDb.addStatusEntry(tupl)
                    #self.statusHashLineList.append(hashLine)

                    cd = bmObject().cyrusDeliveryObj()
                    cd['@timestamp'] = timestp
                    cd['date']       = date + " " +  hour
                    cd['hour']       = hour

                    #ElasticEntries
                    cd['source'] = self.maillog
                    cd['beat']['hostname'] = CFG.HOSTNAME
                    cd['beat']['version']  = CFG.VERSION

                    cd['recip']   = recip
                    cd['domain'] = recip.lower().split('@')[-1].replace(">","")
                    
                    cd['delivery']['msgid'] = messageid
                    cd['delivery']['log']   = thisLine

                    
                    if self.action=="storeToELK":
                        self.s.sendDataStatus(cd)

                    # After post data to ELK 
                    try:
                        del cd
                    except  KeyError:
                        pass

        
        # When user logged to mailbox 
        if self.pushCyrusLoginStatus: 
            cyrusLoginEntries = self.cyrusLogin.match(thisLine)        
            if cyrusLoginEntries is not None:
                hashLine = hashlib.md5(thisLine.encode('ascii')).hexdigest()
                
                # Test if this line is already parsed
                #if hashLine in self.statusHashLineList:
                if self.cDb.verifyIfStatusHashLineExists(hashLine):
                    pass
                else:
                    
                    date      = self.dates[cyrusLoginEntries.group('date')]
                    hour      = cyrusLoginEntries.group('hour')
                    ipAddr    = cyrusLoginEntries.group('ipAddr')
                    user      = cyrusLoginEntries.group('user') 

                    timestp = int(datetime.strptime(date+""+hour,'%Y-%m-%d%H:%M:%S').strftime("%s"))*1000 + self.getDeltaMicroStp(hour)
                    
                    # Feed status table with this new hash for next time bmSearch will re-parsed this this
                    tupl=(hashLine, timestp)
                    self.cDb.addStatusEntry(tupl)
                    #self.statusHashLineList.append(hashLine)

                    cl = bmObject().cyrusLoginObj()
                    cl['@timestamp'] = timestp
                    cl['date']       = date + " " +  hour
                    cl['hour']       = hour

                    #ElasticEntries
                    cl['source'] = self.maillog
                    cl['beat']['hostname'] = CFG.HOSTNAME
                    cl['beat']['version']  = CFG.VERSION

                    cl['user']   = user
                    cl['ipAddr'] = ipAddr
                    cl['log']    = thisLine

                    
                    if self.action=="storeToELK":
                        self.s.sendDataLoggin(cl)

                    # After post data to ELK 
                    try:
                        del cl
                    except  KeyError:
                        pass


        #Because we end the mail bloc log, we can now put it in ELK
        smtpRemovedEntries=self.smtpRemoved.match(thisLine)
        if smtpRemovedEntries is not None:
            queueid=smtpRemovedEntries.group('queueid')

            #1rsttime:create
            if queueid in self.surveyQueuid:
                if self.action=="storeToELK":
                    res = self.s.sendData(self.data[queueid])
                    
                    msg = "[Normal] Post to ELK message referenced with msgid "
                    if 'msgid' in self.data[queueid]:
                        msg += self.data[queueid]['msgid']
                    backoffice.info(msg)
                    
                    
                    
                #In any case we delete all about this message to preserve memory
                try:
                    # In case the msg now delivered while deferred  
                    if queueid in self.bmObjectDSN4:
                        self.bmObjectDSN4.remove(queueid)
                        self.cDb.deleteQueueObj(queueid)
                        
                    if 'msgid' in self.data[queueid]:
                        messageid = self.data[queueid]['msgid']
                        timestamp = self.data[queueid]['@timestamp']/1000
                        
                        tupl = (messageid,timestamp)
                        self.cDb.addEntry(tupl)
                    
                    del self.data[queueid]
                                       
                except  KeyError:
                    pass


            # Delete some bad entries
            for badKey in ('tablehash','warning','statistics','recip','filename'):
                if badKey in self.data:
                    try:
                        del self.data[badKey]
                    except KeyError:
                        pass
                







