#!/usr/bin/env python3
# -*- coding: utf-8; py-indent-offset: 4 -*-


#  +-- bmSearch ------------------------------------------------------------------+
#  |                                                                              |
#  |  _                 _____                          _                          |
#  | | |               / ____|                        | |                         | 
#  | | |___  _ __ ___ | |____    ___  ____   _ _  ___ | |__                       |
#  | |  _  \| '_ ` _ \ \___  \  / _ \|___ \ | `_|/ __||  _  \                     | 
#  | | |_| || | | | | |____|  ||  __// __  || | | |__ | | | |                     |
#  | |_____/|_| |_| |_|______/  \___|\____/ |_|  \___||_| |_|                     |
#  |                                                                              |
#  |                                                                              |
#  +------------------------------------------------------------------------------+
#
# * ----------------------------------------------------------------------------------------*
# * @author  : Pascal SALAUN <pascal.salaun-AT-bm-monitor.org>                              *
# * @Website : http://www.bm-monitor.org <http://www.bm-monitor.org>                        *
# * @license : GNU Affero General Public License Version 3 <http://www.gnu.org/licenses>    *
# *                                                                                         *
# * bmSearch is free software;  you can redistribute it and/or modify it under the          *
# *  terms of the  license mentionned above.                                                *
# * bmSearch is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY    * 
# * ----------------------------------------------------------------------------------------*

# Native Python Lib/Class 
from collections import defaultdict
import json
import requests


# bmSearch Class
import config as CFG
from logger import *
backoffice = logger().setLogger().getChild('pushTemplate')

class pushTemplate:
    """
    pushTemplate:
        Class to PUT bmSearch templates to an ELK platform
    """
    APIURL                 = CFG.APIURL
    loadTemplate           = CFG.loadTemplates
    templateDirPath        = "/etc/bmSearch/tpl/"
    bmSearchTemplates      = [("bmsearchmsg", "bmSearch-elk7.x.template.json"),
                              ("bmsearchstatus", "bmSearchStatus-elk7.x.template.json"),
                              ("bmsearchloggin", "bmSearchLoggin-elk7.x.template.json"),
                              ("bmsearchsynchro","bmSearchSynchro-elk7.x.template.json"),
                              ("bmsearchroutedby","bmSearchRoutedby-elk7.x.template.json"),
                            ]
    
    
    def __init__(self):

        self.logger = logger()
        
        self.params = defaultdict(dict)
        
        self.headers = {'Content-Type': "application/json; charset=utf-8", 'Accept': "application/json", 'Accept-Encoding': 'gzip, deflate'}
        self.params['headers'] = self.headers
        
        if hasattr(CFG,'user') and hasattr(CFG,'password'):
            from requests.auth import HTTPBasicAuth
            self.params['auth'] = HTTPBasicAuth(CFG.user, CFG.password)
        
        if hasattr(CFG,'verifyPeer'):
            self.params['verify'] = CFG.verifyPeer
        
        if hasattr(CFG,'serverCert'):
            self.params['cert'] = CFG.serverCert



    def load(self):
        if self.loadTemplate:
            for tpl in self.bmSearchTemplates:
                self.sendTemplate(*tpl)



    def sendTemplate(self, *args):
        """
            sendData :
                Put data to an ELK platform
                data have to be JSON formated
        """
        
        tplName     = args[0]
        tplFileName = args[1]
        res = False
        
        f = open(self.templateDirPath + tplFileName, "r")
        bmJson = f.read()
        f.close()
        
        url = self.APIURL + "/_template/" + tplName
        self.params['data'] = bmJson

        try:
            res = requests.put(url, **self.params)
            backoffice.debug(res.text)
        except Exception as error:
            backoffice.critical(error)
            pass
